{-# OPTIONS -fglasgow-exts #-}
module Main(main) where

import qualified System
import GenUtil
import GetOpt
import GuidGen
import SimpleParser
import Char
import List(tails)
import Monad
import Time(getClockTime)
import Version



data SourceFile = SourceFile { moduleName :: String, path :: String, guids :: [GUID] }
    deriving(Show)

data Export = Normal String |  Alias String String | Module String
    deriving(Show)
data GUID = GUID String [Export]
    deriving(Show)

data Flag = Verbose | Version | Generate | Output String deriving(Eq)

options :: [OptDescr Flag]
options =
   [Option ['v']     ["verbose"]           (NoArg Verbose)   "verbose mode",
    Option ['V','?'] ["version"]           (NoArg Version)   "show version info",
    Option ['g']     ["generate"]          (NoArg Generate)  "generate a new GUID",
    Option ['o']     ["output"]            (ReqArg Output "DIR")   "directory to output guid modules to"]
    

parseOptions :: IO ([Flag], [String])
parseOptions = do
    argv <- System.getArgs
    case (getOpt Permute options argv) of
        (o,n,[]) | Version `elem` o -> putStrLn (usageInfo header options) >> exitSuccess
        (o,n,[]) -> return (o,n)
        (_,_,errs) -> putErrDie (concat errs ++ usageInfo header options)

header = fullName ++ "\nUsage: hsguid [OPTION...] files..."

    

getOutput s  = head [ o|(Output o) <- s ++ [Output "."] ]

main = do
    (f,ss) <- parseOptions
    let verb = Verbose `elem` f
    if Generate `elem` f then 
        genGuid verb >>= \g -> putStrLn $ "{-# GUID " ++ g ++ "  #-}"
     else case ss of 
        [] -> putStrLn $ usageInfo header options
        ss -> do
            ss <- mapM doFile ss
            let ng = length $ concatMap guids ss
            when verb $ putErrLn $ "Found " ++ nval ng "GUID" ++ " in " ++ nval (length ss)  "file"
            when verb $ putErrLn $ "Outputing files to " ++ getOutput f
            createMods verb (getOutput f) ss

nval 1 s = "1 " ++ s
nval n s = show n ++ " " ++ s ++ "s"

createMods :: Bool -> String -> [SourceFile] -> IO ()
createMods verb odir ss = mapM_ doSource  ss where
    doSource (SourceFile mn pt gs) = mapM_ (doGuid mn pt) gs
    doGuid mn pt (GUID guid es) = do
        when verb $ putErrLn $ unwords ["Processing",  nval (length es) "export", "from", pt, "as",  guid]
        t <- fmap show getClockTime
        let fn = odir ++ "/GUID_" ++ guid ++ ".hs"
        let nm = "{-# OPTIONS -fno-monomorphism-restriction  #-}"
        let bn = unwords ["{- generated by", fullName, "from", pt, "on", t, "-}"]  
        let md = "module GUID_" ++ guid ++ showExports es ++ " where"
        let im = "import " ++ mn
        writeFile fn $ unlines $ [nm, "", bn, md, "", im] ++ getImports es ++ getAliases es
    
showExports es = "(" ++ concatInter ", " (map se es) ++ ")" where 
  se (Normal s) = s
  se (Module m) = "module " ++ m
  se (Alias _ n) = n

getImports es = ["import " ++ mn | (Module mn) <- es]
getAliases es = ["type " ++ a ++ " = " ++ b | Alias b a <- es, isUpper (head b) ] ++
                 [ a ++ " = " ++ b | Alias b a <- es, isLower (head b) ]

ds = dropWhile isSpace
    
doFile :: String -> IO SourceFile
doFile file = readFile file >>= \c -> let 
    ls = map (ds . unlit) $ lines c
    unlit ('>':cs) = cs
    unlit cs = cs
    sfname = head $ concatMap getmod ls
    getmod ('m':'o':'d':'u':'l':'e':c:cs) | isSpace c  = [takeWhile isModname $ ds cs]
    getmod _ = []
    isModname c = isAlphaNum c || c `elem` "_."
    ce (x:xs) = GUID x (map Normal xs)
    gs = concatMap f $ tails c
    f ('{':'-':'#':cs') | ('G':'U':'I':'D':c:cs) <- ds cs', isSpace c, ('-':'#':cs'') <-  reverse (takeWhile (/= '}') (ds cs)) = parser parseGUID $ reverse (ds cs'')
    f _ = []
    in return (SourceFile { moduleName = sfname, path = file, guids = gs} ) 


parseGUID :: Parser GUID
parseGUID = do
    pWS
    guid <- many1 (sat isAlphaNum)
    pWS1
    es <- many (parseExport >>= \e ->  char ',' >> pWS >> return e)
    le <- parseExport 
    parseRest
    return $ GUID guid (es ++ [le])


parseExport = mod <|> r  where
    mod = do
        parseExact "module"
        pWS1 
        n <- parseName 
        return $ Module n
    r = do
        n <- parseName
        ma <- option Nothing $ do
            parseExact "as"
            pWS1
            n2 <- parseName
            return (Just n2)
        case ma of 
            Nothing -> return $ Normal n
            Just n2 -> return $ Alias n n2
        



parseName = do
    n <- many1 (sat (\c -> isAlphaNum c || c `elem` "_." )) 
    v <- option "" $ do
        char '('
        v <- many (sat (/= ')'))
        char ')'
        return $ "(" ++ v ++ ")"
    pWS
    return $ n ++ v

    
pWS :: Parser ()
pWS = whitespace >> (skipOption $ do
    parseExact "--" 
    skipMany (sat (/= '\n'))
    pWS)
    
pWS1 = sat isSpace >> pWS
